import logging
from argparse import ArgumentParser
from pathlib import Path

from inst2ufo.copy import copy_instructions
from inst2ufo.remove import remove_instructions


def copy() -> None:
    parser = ArgumentParser(
        description="Copy TrueType instructions from a binary font to a UFO."
    )
    parser.add_argument(
        "-g",
        "--glyphs",
        action="store_true",
        default=False,
        help="Modify outlines in the UFO to match the TTF exported from Glyphs",
    )
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        nargs=1,
        help="Save UFO to path instead of overwriting the original",
    )
    # parser.add_argument(
    #     "-n",
    #     "--names",
    #     type=str,
    #     nargs=1,
    #     help="Read glyph name mapping from file",
    # )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_argument(
        "ttf",
        type=str,
        nargs=1,
        help="The binary font",
    )
    parser.add_argument(
        "ufo",
        type=str,
        nargs=1,
        help=(
            "The target UFO. If not specified, inst2ufo will look for a UFO with the "
            "same file name stem as the TTF."
        ),
    )
    args = parser.parse_args()
    if args:
        if args.verbose:
            logging.basicConfig(level=logging.INFO)

        ttf_path = Path(args.ttf[0])
        ufo_path = (
            ttf_path.with_suffix(".ufo") if args.ufo is None else Path(args.ufo[0])
        )
        out_path = ufo_path if args.out is None else Path(args.out[0])

        copy_instructions(ttf_path, ufo_path, out_path, fix_glyphs=args.glyphs)

    else:
        parser.print_help()


def remove() -> None:
    parser = ArgumentParser(description="Remove TrueType instructions from a UFO.")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_argument(
        "ufo",
        type=str,
        nargs="+",
        help="The UFO path(s) to remove TrueType instructions from.",
    )
    args = parser.parse_args()
    if args:
        if args.verbose:
            logging.basicConfig(level=logging.INFO)

        for ufo in args.ufo:
            remove_instructions(Path(ufo))

    else:
        parser.print_help()



import logging
from typing import TYPE_CHECKING

from fontTools.pens.hashPointPen import HashPointPen
from fontTools.ttLib import TTFont
from ufoLib2 import Font

from inst2ufo.libkeys import LIB, LIB_CVT, LIB_FPGM, LIB_PREP

if TYPE_CHECKING:
    from pathlib import Path


logger = logging.getLogger(__name__)


def binaryToIntList(value: int, start: int = 0) -> list[int]:
    intList = []
    counter = start
    while value:
        if value & 1:
            intList.append(counter)
        value >>= 1
        counter += 1
    return intList


def copy_cvt(font: TTFont, ufo: Font) -> None:
    if "cvt " not in font:
        logger.info("The font has no control value table (cvt).")
        return
    cvt = font["cvt "]
    logger.info("Copying control value table")
    ufo.lib[LIB][LIB_CVT] = {str(i): v for i, v in enumerate(cvt.values)}


def copy_fpgm(font: TTFont, ufo: Font) -> None:
    if "fpgm" not in font:
        logger.info("The font has no font program table (fpgm).")
        return
    fpgm = font["fpgm"]
    if not hasattr(fpgm, "program"):
        logger.warning("The fpgm table does not contain a program.")
        return
    logger.info("Copying font program table")
    ufo.lib[LIB][LIB_FPGM] = "\n".join(fpgm.program.getAssembly())


def copy_gasp(font: TTFont, ufo: Font) -> None:
    if "gasp" not in font:
        logger.warning("The font has no gasp table.")
        return
    ufo.info.openTypeGaspRangeRecords = [
        {"rangeMaxPPEM": ppm, "rangeGaspBehavior": binaryToIntList(bits)}
        for ppm, bits in font["gasp"].gaspRange.items()
    ]


def copy_glyf(font: TTFont, ufo: Font) -> None:
    if "glyf" not in font:
        logger.warning("The font has no glyf table.")
        return

    glyf = font["glyf"]
    hmtx = font["hmtx"]
    glyph_set = font.getGlyphSet()
    logger.info("Copying glyph programs")
    for glyph_name, pen_glyph in glyph_set.items():
        if glyph_name not in ufo:
            logger.warning(f"Glyph from font not found in UFO: '{glyph_name}'")
            continue
        ufo_glyph = ufo[glyph_name]
        tt_glyph = glyf[glyph_name]
        if LIB in ufo_glyph.lib:
            logger.warning(
                "Overwriting existing TrueType instructions "
                f"in UFO glyph '{glyph_name}'."
            )
        lib = ufo_glyph.lib[LIB] = {}
        if hasattr(tt_glyph, "program"):
            lib["assembly"] = "\n".join(tt_glyph.program.getAssembly())
        else:
            logger.info(f"Glyph '{glyph_name}' does not contain a program.")
            lib["assembly"] = ""
        lib["formatVersion"] = "1"
        width, _ = hmtx[glyph_name]
        hpp = HashPointPen(width, glyph_set)
        pen_glyph.drawPoints(hpp)
        lib["id"] = hpp.hash


def copy_maxp(font: TTFont, ufo: Font) -> None:
    if "maxp" not in font:
        logger.warning("The font has no maximum profile table (maxp).")
        return
    maxp = font["maxp"]
    logger.info("Copying maximum profile table")
    lib = ufo.lib[LIB]
    lib["maxFunctionDefs"] = maxp.maxFunctionDefs
    lib["maxInstructionDefs"] = maxp.maxInstructionDefs
    lib["maxStackElements"] = maxp.maxStackElements
    lib["maxStorage"] = maxp.maxStorage
    lib["maxTwilightPoints"] = maxp.maxTwilightPoints
    lib["maxZones"] = maxp.maxZones


def copy_prep(font: TTFont, ufo: Font) -> None:
    if "prep" not in font:
        logger.info("The font has no control value program table (prep).")
        return
    prep = font["prep"]
    if not hasattr(prep, "program"):
        logger.warning("The prep table does not contain a program.")
        return
    logger.info("Copying control value program table")
    ufo.lib[LIB][LIB_PREP] = "\n".join(prep.program.getAssembly())


def copy_instructions(
    ttf_path: "Path", ufo_path: "Path", out_path: "Path", fix_glyphs: bool = False
) -> None:
    font = TTFont(ttf_path)
    ufo = Font.open(ufo_path)
    logger.info(ufo)

    if LIB in ufo.lib:
        logger.warning("Overwriting existing TrueType instructions in UFO.")

    ufo.lib[LIB] = {}

    copy_cvt(font, ufo)
    copy_gasp(font, ufo)
    copy_fpgm(font, ufo)
    copy_prep(font, ufo)
    copy_maxp(font, ufo)
    copy_glyf(font, ufo)

    if ufo.lib[LIB]:
        ufo.lib[LIB]["formatVersion"] = "1"

    if fix_glyphs:
        from inst2ufo.glyphs import fix_outlines_for_glyphs

        fix_outlines_for_glyphs(ufo)

    ufo.save(out_path, overwrite=True)



from typing import TYPE_CHECKING

from ufoLib2.objects.glyph import Glyph

from inst2ufo.pens import RotateStartPointPen

if TYPE_CHECKING:
    from ufoLib2 import Font


def fix_outlines_for_glyphs(ufo: "Font") -> None:
    # Rotate the outlines so that the start point if on an offcurve, like Glyphs does it
    for ufo_glyph in ufo:
        g = Glyph()
        pp = g.getPointPen()
        rsp = RotateStartPointPen(outPen=pp)
        ufo_glyph.drawPoints(rsp)
        ufo_glyph.clear()
        up = ufo_glyph.getPointPen()
        g.drawPoints(up)



from fontTools.pens.filterPen import ContourFilterPointPen


class RotateStartPointPen(ContourFilterPointPen):
    def filterContour(self, contour):
        contour.reverse()
        # off = contour.pop()
        # contour.insert(0, off)
        return contour



import logging
from typing import TYPE_CHECKING

from ufoLib2 import Font

from inst2ufo.libkeys import LIB

if TYPE_CHECKING:
    from pathlib import Path


logger = logging.getLogger(__name__)


def remove_instructions(ufo_path: "Path") -> None:
    ufo = Font.open(ufo_path)
    logger.info(f"Removing font-level TrueType instructions from '{ufo_path.name}'.")

    if LIB in ufo.lib:
        del ufo.lib[LIB]

    for glyph in ufo:
        if LIB in glyph.lib:
            del glyph.lib[LIB]

    ufo.save(ufo_path, overwrite=True)
